# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 15.11.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Solution proposal for exercise 11, task 7.
"""

# pylint: disable=expression-not-assigned
# pylint: disable=bad-indentation
# pylint: disable=redefined-outer-name
# pylint: disable=import-error

import os

import numpy as np
import matplotlib.pyplot as plt
import cv2
import skimage.morphology as morph
from skimage.feature import peak_local_max
from skimage.color import label2rgb
import scipy.ndimage as ndi

def plot_image(image, fig_num, name=None, colormap='gray', write_file=None,
               fixed_range=True):
  """Plot image"""
  fig = plt.figure(fig_num)
  if fixed_range:
    vmin = np.iinfo(image.dtype).min
    vmax = np.iinfo(image.dtype).max
  else:
    vmin = int(np.min(image))
    vmax = int(np.max(image))
  plt.imshow(image, cmap=colormap, interpolation='none', vmin=vmin, vmax=vmax)
  if name is not None:
    plt.title(name)
  plt.xticks([]), plt.yticks([])
  plt.tight_layout()
  if write_file:
    fig.savefig(write_file, bbox_inches='tight', pad_inches=0)
  fig_num += 1
  return fig_num

def main():
  """Segmentation of cells"""

  print('='*80)
  print('Solution to weekly exercises in INF4300')
  print('Exercise 11')
  print('Task 7')
  print('-'*80)

  # Configurations
  result_dir = 'results/images'
  image_file = '../../images/img_cells.jpg'

  # Step 1: Read image
  image = cv2.imread(image_file, cv2.IMREAD_GRAYSCALE)
  fig_num = plot_image(image, 0,
                       write_file=os.path.join(result_dir, 't07_gray_cells.png'))

  # Step 2: Threshold the image with otsu
  _, thresholded = cv2.threshold(image, 0, 255,
                                 cv2.THRESH_BINARY+cv2.THRESH_OTSU)
  fig_num = plot_image(thresholded, fig_num,
                       write_file=os.path.join(result_dir, 't07_thresholded.png'))

  # Step 3: Fill interior gaps
  filled = morph.remove_small_objects(thresholded.astype(bool), min_size=200)
  fig_num = plot_image(filled.astype(np.uint8)*255, fig_num,
                       write_file=os.path.join(result_dir, 't07_filled.png'))

  # Step 4: Compute distance transform
  distance = ndi.distance_transform_edt((1 - filled).astype(np.uint8))
  fig_num = plot_image(distance, fig_num, fixed_range=False,
                       write_file=os.path.join(result_dir, 't07_filled.png'))

  # Step 5: Compute local maxima of distance
  locmax = peak_local_max(distance, indices=False, footprint=np.ones((3, 3)),
                          labels=~filled)

  # Step 6: Compute watershed
  markers = ndi.label(locmax)[0]
  labels = morph.watershed(-distance.astype(float), markers, mask=~filled)

  # Step 7: Visualize the result
  col_labels = label2rgb(labels)
  fig_num = plot_image(col_labels, fig_num, fixed_range=False,
                       write_file=os.path.join(result_dir, 't07_ws_labels.png'))

  col_labels_merged = label2rgb(labels, image)
  fig_num = plot_image(col_labels_merged, fig_num, fixed_range=False,
                       write_file=os.path.join(result_dir, 't07_ws_label_overlayed.png'))


  plt.show()

if __name__ == "__main__":
  main()
